"""
SSM helper functions
"""
import boto3
from b3d.aws import helpers


def get_parameter(cl: boto3.client, parameter_name: str):
    """
    Describe a parameter, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_parameter, Name=parameter_name
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def delete_parameter(cl: boto3.client, parameter_name: str, dry: bool):
    """
    Delete a parameter
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_parameter, Name=parameter_name
    )



"""
ApiGateway helper functions
"""
import boto3
import b3q
from b3d.aws import helpers


def get_rest_api(cl: boto3.client, api_id: str):
    """
    Describe a rest api, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_rest_api, restApiId=api_id
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def delete_rest_api(cl: boto3.client, api_id: str, dry: bool):
    """
    TODO: create custom waiter for something like RestApiTerminated
     Could just call cl.get_rest_apis() until api_id is not in the
     returned object list
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_rest_api, restApiId=api_id
    )


def get_base_path_mappings(cl: boto3.client, rest_api_id: str, stage_name: [str, None] = None):
    """
    Get all base path mappings associated with some rest api, and optionally a stage within
    that rest api
    """

    mappings = []
    constraints = {
        "restApiId": rest_api_id
    }
    if stage_name is not None:
        constraints["stage"] = stage_name

    domains = get_custom_domain_names(cl)
    for domain in [item["domainName"] for item in domains]:
        for item in b3q.get(
            cl.get_base_path_mappings,
            arguments={"domainName": domain},
            constraints=constraints
        ):
            mappings.append({
                "domain": domain,
                "base_path": item.get("basePath")
            })

    return mappings


@helpers.attempt_api_call_multiple_times
def delete_base_path_mapping(cl: boto3.client, domain_name: str, base_path: str, dry: bool):
    """
    TODO: create custom waiter for BasePathMappingDeleted, could call
     cl.get_base_path_mappings until base_path is not in returned object
     list
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_base_path_mapping, domainName=domain_name, basePath=base_path
    )


def get_usage_plan(cl: boto3.client, usage_plan_id: str):
    """
    Describe a usage plan, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_usage_plan, usagePlanId=usage_plan_id
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def delete_usage_plan(cl: boto3.client, usage_plan_id: str, dry: bool):
    """
    Delete a usage plan
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_usage_plan, usagePlanId=usage_plan_id
    )


@helpers.attempt_api_call_multiple_times
def delete_stage(cl: boto3.client, rest_api_id: str, stage_name: str, dry: bool):
    """
    Delete a stage from some rest api
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_stage, restApiId=rest_api_id, stageName=stage_name
    )


def get_custom_domain_names(cl: boto3.client):
    """
    Retrieve all domain names for some region
    """
    return list(b3q.get(cl.get_domain_names))


def get_api_key(cl: boto3.client, api_key_id: str, include_value: bool = True):
    """
    Describe an api key, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_api_key, apiKey=api_key_id, includeValue=include_value
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def delete_api_key(cl: boto3.client, api_key_id: str, dry: bool):
    """
    Delete an api key
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_api_key, apiKey=api_key_id
    )



"""
Functions that are used to generate different types of log messages
"""


def _new_log_msg(result: str = None, err: str = None, msg: str = None):
    return {
        "result": result,
        "err": err,
        "msg": msg
    }


def log_msg_destroy(resource_type: str, resource_id: str, resp: dict) -> dict:
    """
    Produce a log message that describes the resource being deleted and whether the
    operation was successful
    """

    log_msg = _new_log_msg()

    if resp["ResponseMetadata"]["HTTPStatusCode"] not in [200, 202, 204]:
        log_msg["result"] = "failure"
        log_msg["err"] = resp.get("Error", {})
        prefix = "Unable to delete"
    else:
        log_msg["result"] = "success"
        prefix = "Successfully deleted"
    log_msg["msg"] = f"{prefix} {resource_type} with ID {resource_id}"

    return log_msg


def log_msg_detach(
        resource_type_detached_from: str, resource_type_detached: str,
        resource_id_detached_from: str, resource_id_detached: str,
        resp: dict
) -> dict:
    """
    Produce a log message that describes the resource being detached, the resource it is
    being detached from, and whether the operation was successful
    """

    log_msg = _new_log_msg()
    if resp["ResponseMetadata"]["HTTPStatusCode"] not in [200, 202, 204]:
        log_msg["result"] = "failure"
        log_msg["err"] = resp.get("Error", {})
        prefix = "Unable to detach"
    else:
        log_msg["result"] = "success"
        prefix = "Successfully detached"
    log_msg["msg"] = f"{prefix} {resource_type_detached} with ID {resource_id_detached} " \
                     f"from {resource_type_detached_from} with ID {resource_id_detached_from}"

    return log_msg


def log_msg_disable(resource_type: str, resource_id: str, resp: dict):
    """
    Produce a log message that describes the resource being disabled and whether the operation
    was successful
    """

    log_msg = _new_log_msg()
    if resp["ResponseMetadata"]["HTTPStatusCode"] not in [200, 202, 204]:
        log_msg["result"] = "failure"
        log_msg["err"] = resp.get("Error", {})
        prefix = "Unable to disable"
    else:
        log_msg["result"] = "success"
        prefix = "Successfully disabled"
    log_msg["msg"] = f"{prefix} {resource_type} with ID {resource_id}"

    return log_msg


def log_msg_unsupported_resource(resource_arn: str):
    """
    Produce a log message for a resource that this library does not currently support
    """
    return _new_log_msg(
        result="failure",
        msg=f"Unable to delete resource with ARN {resource_arn} because "
        f"this library currently doesn't support that resource type."
    )



"""
Helper functions for this module
"""
import time
from botocore.exceptions import ClientError
import boto3


def make_call_catch_err(fn: callable, **kwargs):
    """
    Catch ClientError exceptions in AWS API calls
    """

    try:
        return fn(**kwargs)
    except ClientError as ce:
        return ce.response


def wait_on_condition(cl: boto3.client, condition: str, **kwargs):
    """
    Instantiate and run a waiter object
    """

    waiter = cl.get_waiter(condition)
    waiter.wait(**kwargs)


def attempt_api_call_multiple_times(func):
    """
    Sometimes the AWS API needs time to catch up when doing multiple
    delete / detach calls in sequence. This function can be added as
    a decorator for any API call that returns a ["ResponseMetadata"]["HTTPStatusCode"]
    key path in its response dictionary.
    """

    def wrap(*args, **kwargs):

        resp = {}
        attempts = 0

        while attempts < 3:

            resp = func(*args, **kwargs)
            if resp["ResponseMetadata"]["HTTPStatusCode"] in [200, 202, 204]:
                return resp

            attempts += 1
            time.sleep(10)

        return resp
    return wrap


def dry_run_success_resp():
    """
    Produce a generic success response. Used in dry runs
    """
    return {
        "ResponseMetadata": {
            "HTTPStatusCode": 200
        }
    }



"""
Delete procedures for ApiGateway resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class ApiGateway(Service):
    """
    Container class for API Gateway resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "apigateway"

    @staticmethod
    def _delete_base_path_mappings(
            cl: boto3.client, rest_api_id: str, stage_name: [str, None] = None, dry: bool = False
    ) -> List[Dict]:
        """
        Remove all base path mappings associated with either a rest api ID or both a
        rest api ID and some stage name
        """

        resps = []

        # Query all existing base path mappings and filter according to input arguments above
        for bpm in aws.api_gateway.get_base_path_mappings(cl, rest_api_id, stage_name=stage_name):
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="base-path-mapping",
                    resource_id=bpm.get("base_path"),
                    resp=aws.api_gateway.delete_base_path_mapping(
                        cl, domain_name=bpm.get("domain"), base_path=bpm.get("base_path"), dry=dry
                    )
                )
            )

        return resps

    class RestApi(Service.Resource):
        """
        Delete procedure for RestApi objects
        """

        @staticmethod
        def resource_type() -> str:
            return "restapis"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.api_gateway.get_rest_api(
                cl, ApiGateway.RestApi.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("apigateway", region_name=region)
            api_id = ApiGateway.RestApi.extract_resource_id_from_arn(arn)

            if not ApiGateway.RestApi.query(cl, arn):
                return resps

            # Delete all base path mappings associated with this rest api
            resps.extend(ApiGateway._delete_base_path_mappings(cl, api_id, dry=dry))

            # Delete this rest api
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="rest-api",
                    resource_id=api_id,
                    resp=aws.api_gateway.delete_rest_api(
                        cl, api_id, dry
                    )
                )
            )

            return resps

    class UsagePlan(Service.Resource):
        """
        Delete procedure for UsagePlan objects
        """

        @staticmethod
        def resource_type() -> str:
            return "usageplans"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.api_gateway.get_usage_plan(
                cl, ApiGateway.UsagePlan.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _delete_stages(cl: boto3.client, usage_plan_id: str, dry: bool):

            resps = []

            for api_stage in aws.api_gateway.get_usage_plan(cl, usage_plan_id).get("apiStages", []):

                # Delete all base path mappings associated with this stage
                resps.extend(
                    ApiGateway._delete_base_path_mappings(
                        cl, rest_api_id=api_stage.get("apiId"), stage_name=api_stage.get("stage"), dry=dry
                    )
                )

                # Delete this stage
                resps.append(
                    log_msg.log_msg_destroy(
                        resource_type="api-stage",
                        resource_id=api_stage.get("stage"),
                        resp=aws.api_gateway.delete_stage(
                            cl, api_stage.get("apiId"), api_stage.get("stage"), dry
                        )
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("apigateway", region_name=region)
            usage_plan_id = ApiGateway.UsagePlan.extract_resource_id_from_arn(arn)

            if not ApiGateway.UsagePlan.query(cl, arn):
                return resps

            # Delete all stages associated with this usage plan
            resps.extend(ApiGateway.UsagePlan._delete_stages(cl, usage_plan_id, dry))

            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="usage-plan",
                    resource_id=usage_plan_id,
                    resp=aws.api_gateway.delete_usage_plan(cl, usage_plan_id, dry)
                )
            )

            return resps

    class Stage(Service.Resource):
        """
        TODO: This isn't necessary for PoC version of repository, since the UsagePlan resource
            deletes all associated stages. Will want to fill this out going forward though for
            the sake of completeness
        """

        @staticmethod
        def resource_type() -> str:
            return "stages"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return True

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:
            return []

    class ApiKey(Service.Resource):
        """
        Delete procedure for ApiKey objects
        """

        @staticmethod
        def resource_type() -> str:
            return "apikeys"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.api_gateway.get_api_key(
                cl, ApiGateway.ApiKey.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            cl = boto3.client("apigateway", region_name=region)
            api_key_id = ApiGateway.ApiKey.extract_resource_id_from_arn(arn)

            if not ApiGateway.ApiKey.query(cl, arn):
                return []

            return [
                log_msg.log_msg_destroy(
                    resource_type="api-key",
                    resource_id=api_key_id,
                    resp=aws.api_gateway.delete_api_key(cl, api_key_id, dry=dry)
                )
            ]



"""
Module for building a resource-to-delete-procedure map any time this library is invoked
"""
import os
from pkgutil import iter_modules
from importlib import import_module


def list_module(module):
    """
    Extract names of all top-level sub-modules given some parent module
    """
    return [
        name for _, name, is_pkg
        in iter_modules(
            [os.path.dirname(os.path.realpath(module.__file__))]
        )
        if not is_pkg
    ]


def find_objects(modules):
    """
    Extract fully qualified import paths for all sub-modules given some parent module
    """
    ret = set()
    for mod in modules:
        objects = [".".join([mod.__package__, m]) for m in list_module(mod) if mod != "__pycache__"]
        ret.update(objects)
    return sorted(ret)


def extend_resource_map(submodule):
    """
    Produce a dictionary with a single top-level key representing some AWS Service,
    further populated by resources contained within that service. The "submodule" argument
    is some sub-module of the src.b3d.resources module.
    """

    # Import objects from top-level module (e.g. src.b3d.resource.ec2)
    m = import_module(submodule)

    # List all classes included with this module
    module_classes = [
        getattr(m, c) for c in [c for c in dir(m) if c[0] != "_" and c != "Service"]
    ]

    # Extract top-level service type class from this module (for src.b3d.resource.ec2, this is EC2)
    top_level_service_classes = [sc for sc in module_classes if getattr(sc, "service_type", None) is not None]
    if len(top_level_service_classes) != 1:
        raise ValueError("There should be exactly one top level Service class per delete module.")
    top_level_service_class = top_level_service_classes[0]

    # Extract all resource classes from top-level service class
    cc = [
        c for c
        in dir(top_level_service_class)
        if c not in ["service_type", "Resource"] and c[0] != "_"
    ]

    return {
        top_level_service_class.service_type(): {
            getattr(top_level_service_class, c).resource_type(): getattr(top_level_service_class, c) for c in cc
        }
    }


def build_resource_map(module):
    """
    Populate and return a resource map consisting of top-level AWS service types
    (e.g. ec2, iam, etc.), and further keyed on resource types for that service
    (e.g. instances, policies, etc.).
    """

    resource_map = {}
    submodules = find_objects([import_module(module)])
    for sm in submodules:
        resource_map.update(extend_resource_map(sm))
    return resource_map



"""
S3 helper functions
"""
import boto3
from b3d.aws import helpers


def get_bucket(cl: boto3.client, bucket_name: str) -> bool:
    """
    Describe a bucket, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.head_bucket, Bucket=bucket_name
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def delete_bucket(cl: boto3.client, bucket_name: str, dry: bool) -> dict:
    """
    Delete a bucket
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_bucket, Bucket=bucket_name
    )


def get_objects_in_bucket(cl: boto3.client, bucket_name: str) -> list:
    """
    List all objects in a bucket
    """

    resp = helpers.make_call_catch_err(
        cl.list_objects, Bucket=bucket_name
    )
    return [obj.get("Key") for obj in resp.get("Contents", [])]


@helpers.attempt_api_call_multiple_times
def delete_objects(cl: boto3.client, bucket_name: str, objects: list, dry: bool) -> dict:
    """
    Delete specified objects in a bucket
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_objects,
        Bucket=bucket_name,
        Delete={"Objects": [{"Key": k} for k in objects]}
    )



"""
ResourceGroupsTaggingApi helper functions
"""
import boto3
import b3q


def get_resources_by_tag(cl: boto3.client, tag_key: str, tag_value: str):
    """
    Retrieve the ARNs of all resources associated with some tag pair
    """

    response = b3q.get(
        cl.get_resources,
        arguments={"TagFilters": [{"Key": tag_key, "Values": [tag_value]}]},
        attribute="ResourceTagMappingList"
    )
    return [r["ResourceARN"] for r in response]



"""
Delete procedures for KMS resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class KMS(Service):
    """
    Container class for KMS resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "kms"

    class Key(Service.Resource):
        """
        Delete procedure for Key objects
        """

        @staticmethod
        def resource_type() -> str:
            return "key"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:

            key_data = aws.kms.get_key(
                cl, KMS.Key.extract_resource_id_from_arn(resource_arn)
            )

            if key_data is None:
                return False

            return key_data["KeyMetadata"].get("KeyState") != "PendingDeletion"

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("kms", region_name=region)
            key_id = KMS.Key.extract_resource_id_from_arn(arn)

            if not KMS.Key.query(cl, arn):
                return resps

            # Disable this key
            resps.append(
                log_msg.log_msg_disable(
                    resource_type="key",
                    resource_id=key_id,
                    resp=aws.kms.disable_key(cl, key_id, dry)
                )
            )

            # Schedule key for deletion
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="key",
                    resource_id=key_id,
                    resp=aws.kms.schedule_key_deletion(cl, key_id, dry)
                )
            )

            return resps



"""
Boto3 utility library that supports deletion of collections of AWS resources
(such as temporary resources created during unit tests).
"""
from typing import Iterator
import boto3
from b3d import aws, utils


DELETE_PROTOCOL_OBJECT_MAP = utils.build_resource_map("b3d.delete")


def _get_all_resources_with_tag(tag_key: str, tag_value: str, region: str):

    ret = []

    resource_groups_tagging_client = boto3.client("resourcegroupstaggingapi", region_name=region)
    # Retrieve ARNs of all objects with the supplied name/tag pair using the ResourceGroupsTaggingApi.
    # Note that this API does not support *all* AWS resource types, so some manual ARN querying is
    # performed below.
    ret.extend(
        aws.resource_groups_tagging.get_resources_by_tag(
            resource_groups_tagging_client, tag_key, tag_value
        )
    )

    iam_client = boto3.client("iam", region_name=region)
    # Retrieve ARNs of all IAM Users with supplied key/value tag pair
    ret.extend(
        aws.iam.get_all_user_arns_with_tags(iam_client, [(tag_key, tag_value)])
    )
    # Retrieve ARNs of all IAM Roles with supplied key/value tag pair
    ret.extend(
        aws.iam.get_all_role_arns_with_tags(iam_client, [(tag_key, tag_value)])
    )
    # Retrieve ARNs of all IAM Policies with supplied key/value tag pair
    ret.extend(
        aws.iam.get_all_policy_arns_with_tags(iam_client, [(tag_key, tag_value)])
    )

    return ret


def _parse_api_gateway_arn(arn: str):
    """
    Custom parsing rules for API Gateway ARNs
    """

    splt = arn.split(":")
    resource_path = splt[-1].split("/")

    if len(resource_path) > 3 and resource_path[3] == "stages":
        # Form /restapis/<api_id>/stages/<stage_name>
        return "apigateway", "stages"
    # Form /<resource-type>/<resource_id>
    return "apigateway", resource_path[1]


def _parse_arn(arn: str):
    """
    Parse an ARN with the following form:
    'arn:aws:<SERVICE>:<REGION>:<ACCT_ID>:<RESOURCE_TYPE>/<RESOURCE_ID>'
    and return (<SERVICE>, <RESOURCE_TYPE>) tuples for each ARN.
    """

    splt = arn.split(":")

    # API Gateway ARNs have at least one corner case
    if splt[2] == "apigateway":
        return _parse_api_gateway_arn(arn)

    if splt[2] == "s3":
        # The ResourceGroupsTaggingApi only supports S3 buckets and we
        # don't support a manual search-object-with-tag process for other
        # S3 objects at this time
        return "s3", "bucket"

    if splt[2] == "lambda":
        return "lambda", splt[-2]

    return splt[2], splt[-1].split("/")[0]


def _map_arn(arn: str):
    """
    Map an ARN to its corresponding entry in DELETE_PROTOCOL_OBJECT_MAP. If it has no
    corresponding entry (i.e. - this library doesn't currently support that resource
    type), then the UnsupportedResource delete object is returned.
    """

    parsed_arn = _parse_arn(arn)
    try:
        return DELETE_PROTOCOL_OBJECT_MAP[parsed_arn[0]][parsed_arn[1]]
    except KeyError:
        return DELETE_PROTOCOL_OBJECT_MAP["unsupported-service"]["unsupported-resource"]


def _map_arns(arns: list) -> zip:
    """
    Map each ARN to its corresponding delete protocol object.
    """

    mapped_arns = [_map_arn(arn) for arn in arns]
    return zip(arns, mapped_arns)


def delete_resources(tag_key: str, tag_value: str, region: str = "us-east-1", dry=True) -> Iterator[list]:
    """
    Delete all resources with some (<tag_key>, <tag_value>) tag pair, given some region
    """

    # Retrieve ARNs of all objects with the supplied name/tag pair
    resource_arns = _get_all_resources_with_tag(tag_key, tag_value, region)

    # Map each ARN to it's corresponding delete object
    mapped_arns = _map_arns(resource_arns)

    # For each resource, detach it from all dependent objects, delete it, and produce a report
    # of all performed actions, whether they were successful, and error messages for any actions
    # that were unsuccessful.
    for arn, obj in mapped_arns:

        report = obj.destroy(arn, region, dry)
        # Some reports are empty (e.g. KMS keys that have already been scheduled for deletion are
        # picked up by the ResourceGroupsTaggingApi)
        if len(report) > 0:
            yield report



"""
Delete procedures for EC2 resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class EC2(Service):
    """
    Container class for EC2 resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "ec2"

    class Instance(Service.Resource):
        """
        Delete procedure for Instance objects
        """

        @staticmethod
        def resource_type() -> str:
            return "instance"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:

            instance_data = aws.ec2.get_instance(
                cl, EC2.Instance.extract_resource_id_from_arn(resource_arn)
            )

            if instance_data is None:
                return False

            try:
                return instance_data["Reservations"][0]["Instances"][0]["State"]["Name"] != "terminated"
            except IndexError:
                return False

        @staticmethod
        def _detach_all_security_groups(cl: boto3.client, instance_id: str, dry: bool):

            resps = []

            instance_resp = aws.ec2.get_instance(cl, instance_id)
            try:
                all_groups = [
                    sg["GroupId"] for sg in
                    instance_resp["Reservations"][0]["Instances"][0].get("SecurityGroups", [])
                ]
            # Instance terminated but not cleaned up yet
            except IndexError:
                return []

            for sg in all_groups:
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached_from="instance",
                        resource_type_detached="security-group",
                        resource_id_detached_from=instance_id,
                        resource_id_detached=sg,
                        resp=aws.ec2.detach_security_group_from_instance(
                            cl, instance_id, all_groups, sg, dry=dry
                        )
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("ec2", region_name=region)
            instance_id = EC2.Instance.extract_resource_id_from_arn(arn)

            if not EC2.Instance.query(cl, arn):
                return resps

            # Detach all security groups from this instance
            resps.extend(EC2.Instance._detach_all_security_groups(cl, instance_id, dry))

            # Terminate this instance
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="instance",
                    resource_id=instance_id,
                    resp=aws.ec2.delete_instance(cl, instance_id, dry=dry)
                )
            )

            return resps

    class SecurityGroup(Service.Resource):
        """
        Delete procedure for SecurityGroup objects
        """

        @staticmethod
        def resource_type() -> str:
            return "security-group"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.ec2.get_security_group(
                cl, EC2.SecurityGroup.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _detach_from_instances(cl: boto3.client, security_group_id: str, dry: bool) -> List[Dict]:
            """
            Get IDs of all instances this security group is attached to and detach it from each
            """

            resps = []

            # Get IDs of all instances that this security group is attached to
            all_instances_attached = aws.ec2.get_all_instances_with_security_group_id(
                cl, security_group_id
            )

            # Detach this security group from each instance
            for instance_id, group_ids in all_instances_attached:
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached_from="instance",
                        resource_type_detached="security-group",
                        resource_id_detached_from=instance_id,
                        resource_id_detached=security_group_id,
                        resp=aws.ec2.detach_security_group_from_instance(
                            cl, instance_id, group_ids, security_group_id, dry=dry
                        )
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("ec2", region_name=region)
            security_group_id = EC2.SecurityGroup.extract_resource_id_from_arn(arn)

            if not EC2.SecurityGroup.query(cl, arn):
                return resps

            # Detach this security group from instances
            resps.extend(EC2.SecurityGroup._detach_from_instances(cl, security_group_id, dry))

            # Delete this security group
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="security-group",
                    resource_id=security_group_id,
                    resp=aws.ec2.delete_security_group(cl, security_group_id, dry=dry)
                )
            )

            return resps

    class Volume(Service.Resource):
        """
        Delete procedure for Volume objects
        """

        @staticmethod
        def resource_type() -> str:
            return "volume"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.ec2.get_volume(
                cl, EC2.Volume.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _detach_from_instances(cl: boto3.client, volume_id: str, dry: bool) -> List[Dict]:
            """
            Get IDs of all instances this volume is attached to and detach it from each
            """

            resps = []

            # Get IDs of all instances that this volume is attached to
            attachments = aws.ec2.get_volume_attachments(
                cl, volume_id
            )

            # Detach this volume from each instance
            for instance_id in attachments:
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached_from="instance",
                        resource_type_detached="volume",
                        resource_id_detached=volume_id,
                        resource_id_detached_from=instance_id,
                        resp=aws.ec2.detach_volume_from_instance(
                            cl, instance_id, volume_id, dry=dry
                        )
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("ec2", region_name=region)
            volume_id = EC2.Volume.extract_resource_id_from_arn(arn)

            # Abort if this resource doesn't exist
            if not EC2.Volume.query(cl, arn):
                return resps

            # Detach this volume from instances
            resps.extend(EC2.Volume._detach_from_instances(cl, volume_id, dry))

            # Delete this volume
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="volume",
                    resource_id=volume_id,
                    resp=aws.ec2.delete_volume(cl, volume_id, dry=dry)
                )
            )

            return resps



"""
Delete procedures for unsupported resources
"""
import boto3
from b3d.delete import Service
from b3d.utils import log_msg


class UnsupportedService(Service):
    """
    Container class for unsupported resource delete procedures
    """

    @staticmethod
    def service_type():
        return "unsupported-service"

    class UnsupportedResource(Service.Resource):
        """
        Delete procedure for unsupported objects
        """

        @staticmethod
        def resource_type():
            return "unsupported-resource"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return True

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True):
            return log_msg.log_msg_unsupported_resource(arn)



"""
KMS helper functions
"""
import boto3
from b3d.aws import helpers


def get_key(cl: boto3.client, key_id: str):
    """
    Describe a key, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.describe_key, KeyId=key_id
    )

    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def disable_key(cl: boto3.client, key_id: str, dry: bool) -> dict:
    """
    Disable a key
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.disable_key, KeyId=key_id
    )


@helpers.attempt_api_call_multiple_times
def schedule_key_deletion(cl: boto3.client, key_id: str, dry: bool) -> dict:
    """
    Schedule a key for deletion
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.schedule_key_deletion, KeyId=key_id
    )



"""
Abstract class definition for AWS resource delete procedures
"""
import abc
from typing import List, Dict
import boto3


class Service(abc.ABC):
    """
    Abstract container class for AWS resource delete procedures
    """

    @staticmethod
    @abc.abstractmethod
    def service_type() -> str:
        """
        Return the name of this service type.
        """

    class Resource(abc.ABC):
        """
        Abstract class for resource delete procedure
        """

        @staticmethod
        def extract_resource_id_from_arn(arn: str) -> str:
            """
            Extract resource ID from ARN string
            """
            return arn.split("/")[-1]

        @staticmethod
        @abc.abstractmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            """
            Determine if this resource exists.
            """

        @staticmethod
        @abc.abstractmethod
        def resource_type() -> str:
            """
            Return the name of this resource type.
            """

        @staticmethod
        @abc.abstractmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:
            """
            Destroy the resource corresponding to this ARN.
            """



"""
Delete procedures for S3 resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class SSM(Service):
    """
    Container class for SSM resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "ssm"

    class Parameter(Service.Resource):
        """
        Delete procedure for Parameter objects
        """

        @staticmethod
        def resource_type() -> str:
            return "parameter"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.ssm.get_parameter(
                cl, SSM.Parameter.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            cl = boto3.client("ssm", region_name=region)
            parameter_name = SSM.Parameter.extract_resource_id_from_arn(arn)

            if not SSM.Parameter.query(cl, arn):
                return []

            return [
                log_msg.log_msg_destroy(
                    resource_type="parameter",
                    resource_id=parameter_name,
                    resp=aws.ssm.delete_parameter(cl, parameter_name, dry)
                )
            ]



"""
Lambda helper functions
"""
import boto3
from b3d.aws import helpers


@helpers.attempt_api_call_multiple_times
def delete_function(cl: boto3.client, function_name: str, dry: bool):
    """
    Delete a function
    """

    if dry:
        return helpers.dry_run_success_resp()

    delete_resp = helpers.make_call_catch_err(
        cl.delete_function, FunctionName=function_name
    )

    return delete_resp


def get_function(cl: boto3.client, function_name: str):
    """
    Describe a function, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_function, FunctionName=function_name
    )

    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp



"""
EC2 helper functions
"""
import boto3
import b3q
from b3d.aws import helpers


@helpers.attempt_api_call_multiple_times
def delete_instance(cl: boto3.client, instance_id: str, dry: bool):
    """
    Terminate an instance and for it to be fully terminated
    """

    if dry:
        return helpers.dry_run_success_resp()

    resp = helpers.make_call_catch_err(
        cl.terminate_instances, InstanceIds=[instance_id]
    )

    if resp["ResponseMetadata"]["HTTPStatusCode"] == 200:
        helpers.wait_on_condition(cl, "instance_terminated", InstanceIds=[instance_id])

    return resp


def get_instance(cl: boto3.client, instance_id: str):
    """
    Describe an instance, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.describe_instances, InstanceIds=[instance_id]
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def describe_all_instances(cl: boto3.client):
    """
    Describe all instances in some region
    """
    return list(b3q.get(
        cl.describe_instances,
        attribute="Reservations"
    ))


@helpers.attempt_api_call_multiple_times
def delete_security_group(cl: boto3.client, security_group_id: str, dry: bool):
    """
    Delete a security group
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_security_group, GroupId=security_group_id
    )


def get_security_group(cl: boto3.client, security_group_id: str):
    """
    Describe a security group, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.describe_security_groups, GroupIds=[security_group_id]
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def get_default_security_group_id(cl: boto3.client):
    """
    In general, each VPC has a default security group named 'default'. We need
    to retrieve its ID in order to remove security groups from instances, because
    the modify_instance_attribute API call requires at least 1 security group ID
    as part of the request.
    """

    resp = helpers.make_call_catch_err(
        cl.describe_security_groups, GroupNames=["default"]
    )

    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 \
        else resp["SecurityGroups"][0]["GroupId"]


@helpers.attempt_api_call_multiple_times
def delete_volume(cl: boto3.client, volume_id: str, dry: bool):
    """
    Delete a volume
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_volume, VolumeId=volume_id
    )


def get_volume(cl: boto3.client, volume_id: str):
    """
    Describe a volume, if it exists
    """

    resp = helpers.make_call_catch_err(cl.describe_volumes, VolumeIds=[volume_id])
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def get_volume_attachments(cl: boto3.client, volume_id: str):
    """
    Return the IDs for all instances that a given volume is attached to.
    """

    volume_data = get_volume(cl, volume_id)
    if volume_data is None:
        # Volume not found, already deleted.
        return []
    return [
        attachment["InstanceId"]for attachment in volume_data["Volumes"][0]["Attachments"]
    ]


def detach_volume_from_instance(cl: boto3.client, instance_id: str, volume_id: str, dry: bool):
    """
    Detach a volume from a running instance
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.detach_volume, InstanceId=instance_id, VolumeId=volume_id
    )


def detach_security_group_from_instance(
        cl: boto3.client,
        instance_id: str,
        attached_security_groups: list,
        target_security_group_id: str,
        dry: bool
):
    """
    Detach a target security group from an instance.
    """

    if dry:
        return helpers.dry_run_success_resp()

    result_groups = [sg for sg in attached_security_groups if sg != target_security_group_id]
    default_security_group_id = get_default_security_group_id(cl)

    # If this VPC does not have a default security group (which shouldn't happen),
    # we won't be able to detach this security group from an instance because EC2
    # requires that at least 1 security group be attached.
    if default_security_group_id is None and len(result_groups) == 0:
        raise ValueError("No default security group detected for this VPC")

    return helpers.make_call_catch_err(
        cl.modify_instance_attribute,
        InstanceId=instance_id,
        Groups=result_groups if result_groups != [] else [default_security_group_id]
    )


def get_all_instances_with_security_group_id(cl: boto3.client, security_group_id: str):
    """
    Given some security group ID, return the IDs of all instances that its attached to. IDs for
    all security groups currently attached to the instance are also returned, as they are needed
    to detach the target security group later.
    """

    ret = []
    all_instances = describe_all_instances(cl)

    for instance in all_instances:

        # Each instance record has an 'Instances'  list, but there should only be one entry per instance
        instance_data = instance["Instances"][0]
        security_groups = [sg["GroupId"] for sg in instance_data.get("SecurityGroups", [])]
        if security_group_id in security_groups:
            ret.append(
                (
                    instance_data.get("InstanceId"),
                    [gid for gid in security_groups if gid != security_group_id]
                )
            )

    return ret



"""
Delete procedures for IAM resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class IAM(Service):
    """
    Container class for IAM resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "iam"

    class User(Service.Resource):
        """
        Delete procedure for User objects
        """

        @staticmethod
        def resource_type() -> str:
            return "user"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.iam.get_user(
                cl, IAM.User.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _delete_permissions_boundary(cl: boto3.client, user_name: str, dry: bool) -> Dict:
            return log_msg.log_msg_detach(
                resource_type_detached="permissions-boundary",
                resource_type_detached_from="user",
                resource_id_detached="N/A",
                resource_id_detached_from=user_name,
                resp=aws.iam.delete_user_permissions_boundary(cl, user_name, dry)
            )

        @staticmethod
        def _delete_all_inline_policies(cl: boto3.client, user_name: str, dry: bool) -> List[Dict]:

            resps = []
            all_inline_policies = aws.iam.get_inline_user_policies(cl, user_name)
            for ip in all_inline_policies:
                resps.append(
                    log_msg.log_msg_destroy(
                        resource_type="inline-policy",
                        resource_id=ip,
                        resp=aws.iam.delete_inline_user_policy(cl, user_name, ip, dry)
                    )
                )

            return resps

        @staticmethod
        def _detach_all_policies(cl: boto3.client, user_name: str, dry: bool) -> List[Dict]:

            resps = []
            all_policies_attached = aws.iam.get_attached_user_policies(cl, user_name)
            for p in all_policies_attached:
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached="policy",
                        resource_type_detached_from="user",
                        resource_id_detached=p.get("PolicyArn", "N/A"),
                        resource_id_detached_from=user_name,
                        resp=aws.iam.detach_policy_from_user(cl, user_name, p.get("PolicyArn", ""), dry)
                    )
                )

            return resps

        @staticmethod
        def _detach_all_access_keys(cl: boto3.client, user_name: str, dry: bool) -> List[Dict]:

            resps = []
            all_access_keys_attached = aws.iam.get_user_access_keys(cl, user_name)
            for ak in all_access_keys_attached:

                delete_resp = aws.iam.delete_access_key(cl, user_name, ak.get("AccessKeyId", ""), dry)
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached="access-key",
                        resource_type_detached_from="user",
                        resource_id_detached=ak.get("AccessKeyId", "N/A"),
                        resource_id_detached_from=user_name,
                        resp=delete_resp
                    )
                )
                resps.append(
                    log_msg.log_msg_destroy(
                        resource_type="access-key",
                        resource_id=ak.get("AccessKeyId", "N/A"),
                        resp=delete_resp
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("iam", region_name=region)
            user_name = IAM.User.extract_resource_id_from_arn(arn)

            # Abort if this resource doesn't exist
            if not IAM.User.query(cl, arn):
                return resps

            # Remove permissions boundary for this user, if one exists
            if aws.iam.user_has_permissions_boundary(cl, user_name):
                resps.append(IAM.User._delete_permissions_boundary(cl, user_name, dry=dry))

            # Delete all inline policies attached to this user
            resps.extend(IAM.User._delete_all_inline_policies(cl, user_name, dry))

            # Detach all policies attached to this user
            resps.extend(IAM.User._detach_all_policies(cl, user_name, dry))

            # Detach and delete all access keys associated with this user
            resps.extend(IAM.User._detach_all_access_keys(cl, user_name, dry))

            # Delete this user
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="user",
                    resource_id=arn,
                    resp=aws.iam.delete_user(cl, user_name, dry)
                )
            )

            return resps

    class Policy(Service.Resource):
        """
        Delete procedure for Policy objects
        """

        @staticmethod
        def resource_type() -> str:
            return "policy"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.iam.get_policy(cl, resource_arn) is not None

        @staticmethod
        def _detach_from_users(cl: boto3.client, policy_arn: str, user_names: list, dry: bool) -> List[Dict]:
            """
            For each user that this policy is associated with, determine whether the policy is being used
            as a permissions boundary for that user or if it is simply attached. In the former case, the
            permissions boundary for that user is destroyed (since there is only ever one). In the latter
            case, the policy is detached from the user.
            """

            resps = []
            for user in user_names:
                if aws.iam.policy_is_permissions_boundary_for_user(cl, user, policy_arn):
                    resps.append(
                        log_msg.log_msg_detach(
                            resource_type_detached="permissions-boundary",
                            resource_type_detached_from="user",
                            resource_id_detached=policy_arn,
                            resource_id_detached_from=user,
                            resp=aws.iam.delete_user_permissions_boundary(cl, user, dry)
                        )
                    )
                else:
                    resps.append(
                        log_msg.log_msg_detach(
                            resource_type_detached="policy",
                            resource_id_detached=IAM.Policy.extract_resource_id_from_arn(policy_arn),
                            resource_type_detached_from="user",
                            resource_id_detached_from=user,
                            resp=aws.iam.detach_policy_from_user(cl, user, policy_arn, dry)
                        )
                    )

            return resps

        @staticmethod
        def _detach_from_groups(cl: boto3.client, policy_arn: str, group_names: list, dry: bool) -> List[Dict]:

            resps = []
            for group in group_names:
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached="policy",
                        resource_id_detached=IAM.Policy.extract_resource_id_from_arn(policy_arn),
                        resource_type_detached_from="group",
                        resource_id_detached_from=group,
                        resp=aws.iam.detach_policy_from_group(cl, group, policy_arn, dry)
                    )
                )

            return resps

        @staticmethod
        def _detach_from_roles(cl: boto3.client, policy_arn: str, role_names: list, dry: bool) -> List[Dict]:

            resps = []
            for role in role_names:
                if aws.iam.policy_is_permissions_boundary_for_role(cl, role, policy_arn):
                    resps.append(
                        log_msg.log_msg_detach(
                            resource_type_detached="permissions-boundary",
                            resource_type_detached_from="role",
                            resource_id_detached=policy_arn,
                            resource_id_detached_from=role,
                            resp=aws.iam.delete_role_permissions_boundary(cl, role, dry)
                        )
                    )
                else:
                    resps.append(
                        log_msg.log_msg_detach(
                            resource_type_detached="policy",
                            resource_id_detached=IAM.Policy.extract_resource_id_from_arn(policy_arn),
                            resource_type_detached_from="role",
                            resource_id_detached_from=role,
                            resp=aws.iam.detach_role_policy(cl, role, policy_arn, dry)
                        )
                    )

            return resps

        @staticmethod
        def _delete_non_default_versions(cl: boto3.client, policy_arn: str, dry: bool) -> List[Dict]:

            resps = []
            policy_versions = aws.iam.list_policy_versions(cl, policy_arn)
            for version in policy_versions.get("Versions", []):

                # Can only delete default policy version via different API call
                if not version.get("IsDefaultVersion"):
                    vid = version.get("VersionId")
                    resps.append(
                        log_msg.log_msg_destroy(
                            resource_type="policy-version",
                            resource_id=f"{IAM.Policy.extract_resource_id_from_arn(policy_arn)}-{vid}",
                            resp=aws.iam.delete_policy_version(cl, policy_arn, vid, dry)
                        )
                    )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("iam", region_name=region)

            # Abort if this resource doesn't exist
            if not IAM.Policy.query(cl, arn):
                return resps

            # Get IDs of all Users, Groups, and Roles that this policy is attached to
            entities_attached = aws.iam.list_entities_policy_attached(cl, arn)

            # Detach this policy from users
            resps.extend(
                IAM.Policy._detach_from_users(
                    cl, arn, [u.get("UserName") for u in entities_attached.get("PolicyUsers", [])], dry
                )
            )

            # Detach this policy from groups
            resps.extend(
                IAM.Policy._detach_from_groups(
                    cl, arn, [g.get("GroupName") for g in entities_attached.get("PolicyGroups", [])], dry
                )
            )

            # Detach this policy from roles
            resps.extend(
                IAM.Policy._detach_from_roles(
                    cl, arn, [r.get("RoleName") for r in entities_attached.get("PolicyRoles", [])], dry
                )
            )

            resps.extend(
                IAM.Policy._delete_non_default_versions(cl, arn, dry)
            )

            # Delete default policy version
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="policy",
                    resource_id=IAM.Policy.extract_resource_id_from_arn(arn),
                    resp=aws.iam.delete_policy(cl, arn, dry)
                )
            )

            return resps

    class Role(Service.Resource):
        """
        Delete procedure for Role objects
        """

        @staticmethod
        def resource_type() -> str:
            return "role"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.iam.get_role(
                cl, IAM.Role.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _remove_permissions_boundary(cl: boto3.client, role_name: str, dry: bool) -> dict:
            return log_msg.log_msg_destroy(
                resource_type="role-permissions-boundary",
                resource_id=f"permissions-boundary-{role_name}",
                resp=aws.iam.delete_role_permissions_boundary(cl, role_name, dry)
            )

        @staticmethod
        def _delete_embedded_policies(cl: boto3.client, role_name: str, dry: bool) -> List[Dict]:

            resps = []
            embedded_policies = aws.iam.list_embedded_role_policies(cl, role_name)
            for policy_name in embedded_policies.get("PolicyNames", []):
                resps.append(
                    log_msg.log_msg_destroy(
                        resource_type="embedded-policy",
                        resource_id=policy_name,
                        resp=aws.iam.delete_role_policy(cl, role_name, policy_name, dry)
                    )
                )

            return resps

        @staticmethod
        def _detach_policies(cl: boto3.client, role_name: str, dry: bool) -> List[Dict]:

            resps = []
            attached_policies = aws.iam.list_attached_role_policies(cl, role_name)
            for policy in attached_policies.get("AttachedPolicies", []):
                resps.append(
                    log_msg.log_msg_detach(
                        resource_type_detached_from="role",
                        resource_type_detached="policy",
                        resource_id_detached_from=role_name,
                        resource_id_detached=policy.get("PolicyArn"),
                        resp=aws.iam.detach_role_policy(cl, role_name, policy.get("PolicyArn"), dry)
                    )
                )

            return resps

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("iam", region_name=region)
            role_name = IAM.Role.extract_resource_id_from_arn(arn)

            # Abort if this resource doesn't exist
            if not IAM.Role.query(cl, arn):
                return resps

            # Remove permissions boundary from this role, if it exists
            if aws.iam.role_has_permissions_boundary(cl, role_name):
                resps.append(IAM.Role._remove_permissions_boundary(cl, role_name, dry))

            # If this role has any embedded policies, delete them
            resps.extend(IAM.Role._delete_embedded_policies(cl, role_name, dry))

            # If this role has any policies attached, detach them
            resps.extend(IAM.Role._detach_policies(cl, role_name, dry))

            # Delete this role
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="role",
                    resource_id=role_name,
                    resp=aws.iam.delete_role(cl, role_name, dry)
                )
            )

            return resps



"""
Delete procedures for Lambda resources
"""
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class Lambda(Service):
    """
    Container class for Lambda resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "lambda"

    class Function(Service.Resource):
        """
        Delete procedure for Function objects
        """

        @staticmethod
        def extract_resource_id_from_arn(arn: str) -> str:
            return arn.split(":")[-1]

        @staticmethod
        def resource_type() -> str:
            return "function"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.lambda_.get_function(
                cl, Lambda.Function.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("lambda", region_name=region)
            function_name = Lambda.Function.extract_resource_id_from_arn(arn)

            if not Lambda.Function.query(cl, arn):
                return resps

            # Destroy this function
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="function",
                    resource_id=function_name,
                    resp=aws.lambda_.delete_function(cl, function_name, dry=dry)
                )
            )

            return resps



"""
IAM helper functions
"""
from typing import List, Dict, Tuple
import boto3
import b3q
from b3d.aws import helpers


@helpers.attempt_api_call_multiple_times
def delete_policy(cl: boto3.client, policy_arn: str, dry: bool) -> dict:
    """
    Delete a policy
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_policy, PolicyArn=policy_arn
    )


def get_policy(cl: boto3.client, policy_arn: str) -> dict:
    """
    Describe a policy, if it exists
    """

    resp = helpers.make_call_catch_err(cl.get_policy, PolicyArn=policy_arn)
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def get_all_policies(cl: boto3.client, scope: str = "Local", prefix: str = "/"):
    """
    Get all policies in some region
    """
    return list(b3q.get(
        cl.list_policies, attribute="Policies", arguments={"PathPrefix": prefix, "Scope": scope}
    ))


def policy_has_tags(cl: boto3.client, policy_arn: str, tags: List[Tuple]) -> bool:
    """
    Determine whether a policy is associated with any in some list of tag pairs
    """

    policy_data = get_policy(cl, policy_arn)
    if policy_data is None:
        return False

    for tag_key, tag_value in tags:
        for t in policy_data["Policy"].get("Tags", []):

            if t["Key"] == tag_key and t["Value"] == tag_value:
                return True

    return False


def get_all_policy_arns_with_tags(cl: boto3.client, tags: List[Tuple]) -> List[str]:
    """
    Get all policies associated with this AWS account that at least one tag from <tags>
    is attached to
    """

    ret = []
    all_policies = get_all_policies(cl)
    for p in all_policies:

        policy_arn = p.get("Arn", None)
        if policy_arn is not None:
            if policy_has_tags(cl, policy_arn, tags):
                ret.append(policy_arn)

    return ret


def list_entities_policy_attached(cl: boto3.client, policy_arn: str) -> dict:
    """
    List all resources that are associated with some policy ARN
    """
    return helpers.make_call_catch_err(
        cl.list_entities_for_policy, PolicyArn=policy_arn
    )


def list_policy_versions(cl: boto3.client, policy_arn: str) -> dict:
    """
    List all versions for a policy
    """
    return helpers.make_call_catch_err(
        cl.list_policy_versions, PolicyArn=policy_arn
    )


@helpers.attempt_api_call_multiple_times
def delete_policy_version(cl: boto3.client, policy_arn: str, version_id: str, dry: bool) -> dict:
    """
    Delete a version of a policy
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_policy_version, PolicyArn=policy_arn, VersionId=version_id
    )


@helpers.attempt_api_call_multiple_times
def delete_role(cl: boto3.client, role_name: str, dry: bool) -> dict:
    """
    Delete a role
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_role, RoleName=role_name
    )


@helpers.attempt_api_call_multiple_times
def delete_role_permissions_boundary(cl: boto3.client, role_name: str, dry: bool) -> dict:
    """
    Delete the permissions boundary assocatied with some role, if it exists
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_role_permissions_boundary, RoleName=role_name
    )


@helpers.attempt_api_call_multiple_times
def delete_role_policy(cl: boto3.client, role_name: str, policy_name: str, dry: bool) -> dict:
    """
    Delete an inline policy from a role
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_role_policy, RoleName=role_name, PolicyName=policy_name
    )


@helpers.attempt_api_call_multiple_times
def detach_role_policy(cl: boto3.client, role_name: str, policy_arn: str, dry: bool) -> dict:
    """
    Detach a policy from a role
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.detach_role_policy, RoleName=role_name, PolicyArn=policy_arn
    )


def get_role(cl: boto3.client, role_name: str) -> dict:
    """
    Describe a role, if it exists
    """

    resp = helpers.make_call_catch_err(cl.get_role, RoleName=role_name)
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def get_all_roles(cl: boto3.client, prefix: str = "/") -> List[Dict]:
    """
    List all roles associated with this AWS account
    """
    return list(b3q.get(
        cl.list_roles, attribute="Roles", arguments={"PathPrefix": prefix}
    ))


def role_has_permissions_boundary(cl: boto3.client, role_name: str) -> bool:
    """
    Determine whether a role has a permissions boundary
    """

    role_data = get_role(cl, role_name)
    if role_data is not None:
        # If this role has a nonempty permissions boundary, there will be a
        # "PermissionsBoundary" key present in its object record
        return role_data["Role"].get("PermissionsBoundary") is not None

    return False


def list_embedded_role_policies(cl: boto3.client, role_name: str):
    """
    List all inline policies for a role
    """
    return helpers.make_call_catch_err(
        cl.list_role_policies, RoleName=role_name
    )


def list_attached_role_policies(cl: boto3.client, role_name: str):
    """
    List all attached policies for a role
    """
    return helpers.make_call_catch_err(
        cl.list_attached_role_policies, RoleName=role_name
    )


def role_has_tags(cl: boto3.client, role_name: str, tags: List[Tuple]) -> bool:
    """
    Determine whether any in some list of tag key / value pairs is associated with
    some role
    """

    role_data = get_role(cl, role_name)
    if role_data is None:
        return False

    for tag_key, tag_value in tags:
        for t in role_data["Role"].get("Tags", []):

            # For Role resources, tag key is always a single string (not a list)
            if t["Key"] == tag_key and t["Value"] == tag_value:
                return True

    return False


def get_all_role_arns_with_tags(cl: boto3.client, tags: List[Tuple]) -> List[str]:
    """
    Return the ARNs of all roles that have at least one tag from the tags list
    """

    ret = []
    all_roles = get_all_roles(cl)
    for r in all_roles:

        role_name = r.get("RoleName", None)
        if role_name is not None:
            if role_has_tags(cl, role_name, tags):
                ret.append(r.get("Arn"))

    return ret


@helpers.attempt_api_call_multiple_times
def delete_instance_profile(cl: boto3.client, instance_profile_name: str, dry: bool) -> dict:
    """
    Delete an instance profile
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_instance_profile, InstanceProfileName=instance_profile_name
    )


@helpers.attempt_api_call_multiple_times
def delete_access_key(cl: boto3.client, user_name: str, access_key_id: str, dry: bool) -> dict:
    """
    Delete an access key pair associated with an IAM user
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_access_key, UserName=user_name, AccessKeyId=access_key_id
    )


def get_instance_profile(cl: boto3.client, instance_profile_name: str) -> dict:
    """
    Describe an instance profile, if it exists
    """

    resp = helpers.make_call_catch_err(cl.get_instance_profile, InstanceProfileName=instance_profile_name)
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


def get_attached_user_policies(cl: boto3.client, user_name: str) -> List[Dict]:
    """
    List all policies attached to a user
    """

    resp = helpers.make_call_catch_err(
        cl.list_attached_user_policies, UserName=user_name
    )
    return [] if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp.get("AttachedPolicies", [])


def get_inline_user_policies(cl: boto3.client, user_name: str) -> List[str]:
    """
    List all inline policies attached to a user
    """

    resp = helpers.make_call_catch_err(
        cl.list_user_policies, UserName=user_name
    )
    return [] if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp.get("PolicyNames", [])


def get_user_access_keys(cl: boto3.client, user_name: str) -> List[Dict]:
    """
    List all access keys associated with a user
    """
    resp = helpers.make_call_catch_err(
        cl.list_access_keys, UserName=user_name
    )
    return [] if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp.get("AccessKeyMetadata", [])


def get_all_users(cl: boto3.client) -> List[Dict]:
    """
    List all users for this AWS account
    """
    return list(b3q.get(
        cl.list_users, attribute="Users"
    ))


@helpers.attempt_api_call_multiple_times
def delete_user(cl: boto3.client, user_name: str, dry: bool) -> dict:
    """
    Delete a user
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_user, UserName=user_name
    )


def get_user(cl: boto3.client, user_name: str) -> dict:
    """
    Describe a user, if it exists
    """

    resp = helpers.make_call_catch_err(
        cl.get_user, UserName=user_name
    )
    return None if resp["ResponseMetadata"]["HTTPStatusCode"] != 200 else resp


@helpers.attempt_api_call_multiple_times
def detach_policy_from_user(cl: boto3.client, user_name: str, policy_arn: str, dry: bool):
    """
    Detach a policy from a user
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.detach_user_policy, UserName=user_name, PolicyArn=policy_arn
    )


@helpers.attempt_api_call_multiple_times
def delete_inline_user_policy(cl: boto3.client, user_name: str, policy_name: str, dry: bool) -> dict:
    """
    Delete an inline policy for some user
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_user_policy, UserName=user_name, PolicyName=policy_name
    )


def user_has_permissions_boundary(cl: boto3.client, user_name: str) -> bool:
    """
    Determine whether a permissions boundary exists for a user
    """

    user_data = get_user(cl, user_name)

    # If user doesn't exist, return False
    if user_data["ResponseMetadata"]["HTTPStatusCode"] != 200:
        return False

    # If user exists and they have a permissions boundary set, return True
    return user_data["User"].get("PermissionsBoundary") is not None


def policy_is_permissions_boundary_for_user(cl: boto3.client, user_name: str, policy_arn: str) -> bool:
    """
    Given some user name and a policy ARN, determine whether that policy ARN is being used as the
    permissions boundary for that user.
    """

    user_data = get_user(cl, user_name)

    # If user doesn't exist, return False
    if user_data["ResponseMetadata"]["HTTPStatusCode"] != 200:
        return False

    # If user exists and they have a permissions boundary set, compare it to policy_arn
    if user_data["User"].get("PermissionsBoundary") is not None:
        return user_data["User"]["PermissionsBoundary"].get("PermissionsBoundaryArn") == policy_arn

    # If user exists and no permissions boundary is set, return False
    return False


def policy_is_permissions_boundary_for_role(cl: boto3.client, role_name: str, policy_arn: str) -> bool:
    """
    Given some role name and a policy ARN, determine whether that policy ARN is being used as the
    permissions boundary for that role.
    """

    role_data = get_role(cl, role_name)

    # If role doesn't exist, return False
    if role_data["ResponseMetadata"]["HTTPStatusCode"] != 200:
        return False

    # If role exists and it has a permissions boundary set, compare it to policy_arn
    if role_data["Role"].get("PermissionsBoundary") is not None:
        return role_data["Role"]["PermissionsBoundary"].get("PermissionsBoundaryArn") == policy_arn

    # If role exists and no permissions boundary is set, return False
    return False


@helpers.attempt_api_call_multiple_times
def delete_user_permissions_boundary(cl: boto3.client, user_name: str, dry: bool) -> dict:
    """
    Delete a permissions boundary from a user
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.delete_user_permissions_boundary, UserName=user_name
    )


def user_has_tags(cl: boto3.client, user_name: str, tags: List[Tuple]) -> bool:
    """
    Determine whether any in some list of tag key / value pairs is associated with
    some user profile
    """

    user_data = get_user(cl, user_name)
    if user_data is None:
        return False

    for tag_key, tag_value in tags:
        for t in user_data["User"].get("Tags", []):

            # For User resources, tag key is always a single string (not a list)
            if t["Key"] == tag_key and t["Value"] == tag_value:
                return True

    return False


def get_all_user_arns_with_tags(cl: boto3.client, tags: List[Tuple]) -> List[str]:
    """
    Return the ARNs of all users that have at least one tag from the tags list
    """

    ret = []
    all_users = get_all_users(cl)
    for u in all_users:

        user_name = u.get("UserName", None)
        if user_name is not None:
            if user_has_tags(cl, user_name, tags):
                ret.append(u.get("Arn"))

    return ret


@helpers.attempt_api_call_multiple_times
def detach_policy_from_group(cl: boto3.client, group_name: str, policy_arn: str, dry: bool):
    """
    Detach a policy from a group
    """

    if dry:
        return helpers.dry_run_success_resp()

    return helpers.make_call_catch_err(
        cl.detach_group_policy, GroupName=group_name, PolicyArn=policy_arn
    )



""" Delete procedures for S3 resources """
from typing import List, Dict
import boto3
from b3d.delete import Service
from b3d import aws
from b3d.utils import log_msg


class S3(Service):
    """
    Container class for S3 resource delete procedures
    """

    @staticmethod
    def service_type() -> str:
        return "s3"

    class Bucket(Service.Resource):
        """
        Delete procedure for Bucket objects
        """

        @staticmethod
        def extract_resource_id_from_arn(arn: str) -> str:
            return arn.split(":")[-1]

        @staticmethod
        def resource_type() -> str:
            return "bucket"

        @staticmethod
        def query(cl: boto3.client, resource_arn: str) -> bool:
            return aws.s3.get_bucket(
                cl, S3.Bucket.extract_resource_id_from_arn(resource_arn)
            ) is not None

        @staticmethod
        def _delete_objects(cl: boto3.client, bucket_name: str, objects: list, dry: bool) -> dict:
            return log_msg.log_msg_destroy(
                resource_type="objects",
                resource_id=" | ".join(objects),
                resp=aws.s3.delete_objects(cl, bucket_name, objects, dry)
            )

        @staticmethod
        def destroy(arn: str, region: str, dry: bool = True) -> List[Dict]:

            resps = []
            cl = boto3.client("s3", region_name=region)
            bucket_name = S3.Bucket.extract_resource_id_from_arn(arn)

            if not S3.Bucket.query(cl, arn):
                return resps

            # Remove all objects stored in this bucket
            objects = aws.s3.get_objects_in_bucket(cl, bucket_name)
            if len(objects) > 0:
                resps.append(S3.Bucket._delete_objects(cl, bucket_name, objects, dry))

            # Delete this bucket
            resps.append(
                log_msg.log_msg_destroy(
                    resource_type="bucket",
                    resource_id=bucket_name,
                    resp=aws.s3.delete_bucket(cl, bucket_name, dry)
                )
            )

            return resps
